unit SXSkinControl;

////////////////////////////////////////////////////////////////////////////////
// SXSkinComponents: Skinnable Visual Controls with Alpha Channel             //
//----------------------------------------------------------------------------//
// Version: 1.0.1                                                             //
// Author: Alexey Sadovnikov                                                  //
// Web Site: http://www.saarixx.info                                          //
// E-Mail: mail@saarixx.info                                                  //
//----------------------------------------------------------------------------//
// LICENSE:                                                                   //
// 1. You can freely distribute this file.                                    //
// 2. You can not make any changes to this file.                              //
// 3. The only person who can change this file is Alexey Sadovnikov.          //
// 4. You can use this file in your freeware projects.                        //
// 5. If you want to use this file in your shareware project, you should      //
//    purchase a project license or a personal license of SXSkinComponents.   //
// 6. You can freely use, distribute and modify skins for SXSkinComponents.   //
// 7. You can create skins for SXSkinComponents.                              //
//----------------------------------------------------------------------------//
// Copyright 2006 by Alexey Sadovnikov. All Rights Reserved.                  //
////////////////////////////////////////////////////////////////////////////////

interface

{$I Compilers.inc}

uses GR32_Image, GR32, Windows, Graphics, Classes, Messages, Forms, StdCtrls,
     ExtCtrls, SysUtils, Controls, SXSkinLibrary;

type

  TSXSkinCustomControl=class(TCustomControl)
   private
    FSkinLibrary:TSXSkinLibrary;
    //FHintData:TSXHintData;
    FOnMouseDown:TMouseEvent;
    FOnMouseEnter:TNotifyEvent;
    FOnMouseLeave:TNotifyEvent;
    FOnMouseMove:TMouseMoveEvent;
    FOnMouseUp:TMouseEvent;
    DrawBR:TRect;
    DrawCR:TRect;
    DrawRgn:HRGN;
    FSkinStyle:String;
    procedure SetSkinStyle(const Value:String);
    procedure SetSkinLibrary(Value:TSXSkinLibrary);
    procedure WMPaint(var Msg:TWMPaint); message WM_PAINT;
    procedure WMEraseBkgnd(var Msg: TWmEraseBkgnd); message WM_ERASEBKGND;
   protected
    procedure SetParent(AParent:TWinControl); override;
    procedure Notification(AComponent:TComponent;Operation:TOperation); override;
    procedure CreateParams(var Params:TCreateParams); override;
    function NeedToPaintBackground:Boolean; virtual;
    procedure Paint; override;
    procedure CMMouseEnter(var Msg:TMessage); message CM_MOUSEENTER;
    procedure CMMouseLeave(var Msg:TMessage); message CM_MOUSELEAVE;
    //procedure CMHintShow(var Message:TCMHintShow); message CM_HINTSHOW;
    procedure MouseDown(Button:TMouseButton;Shift:TShiftState;X,Y:Integer); override;
    procedure MouseMove(Shift:TShiftState;X,Y:Integer); override;
    procedure MouseUp(Button:TMouseButton;Shift:TShiftState;X,Y:Integer); override;
   public
    procedure PaintRectToBitmap(DestCanvasHandle:HDC;DestCanvasRect:TRect;
               Rect:TRect;Rgn:HRGN;Bitmap:TBitmap32;X,Y:Integer;
               WithSubItems:Boolean); virtual;   
    procedure SkinChanged; virtual;
    function IsTransparent(X,Y:Integer;Limit:Integer=10):Boolean; virtual;
    function CanShowControl:Boolean; virtual;
    procedure SetBounds(ALeft,ATop,AWidth,AHeight:Integer); override;
    constructor Create(AOwner:TComponent); override;
    destructor Destroy; override;
    property Canvas;
    //property HintData:TSXHintData read FHintData write FHintData;
    property OnMouseDown:TMouseEvent read FOnMouseDown write FOnMouseDown;
    property OnMouseEnter:TNotifyEvent read FOnMouseEnter write FOnMouseEnter;
    property OnMouseLeave:TNotifyEvent read FOnMouseLeave write FOnMouseLeave;
    property OnMouseMove:TMouseMoveEvent read FOnMouseMove write FOnMouseMove;
    property OnMouseUp:TMouseEvent read FOnMouseUp write FOnMouseUp;
    property SkinLibrary:TSXSkinLibrary read FSkinLibrary write SetSkinLibrary;
    property SkinStyle:String read FSkinStyle write SetSkinStyle;
   published
    property Left default 0;
    property Top default 0;
  end;

 TGetWindowRgnBox=function(hWnd:HWND;var Rect:TRect):Integer; stdcall;

var  TestingRegions:Boolean=False;
 ControlsNotToPaint:TList;
   CanvasNotToPaint:TList;
   FGetWindowRgnBox:TGetWindowRgnBox;

implementation

uses SXSkinPanel;

procedure InitGetWindowRgnBox;
var HDLL:THandle;
begin
 FGetWindowRgnBox:=nil;
 HDLL:=LoadLibrary('user32.dll');
 FGetWindowRgnBox:=GetProcAddress(HDLL,'GetWindowRgnBox');
end;

function DoShowControl(Control:TControl):Boolean;
begin
 if Control is TSXSkinCustomControl then
  Result:=TSXSkinCustomControl(Control).CanShowControl else
   Result:=Control.Visible or (csDesigning in Control.ComponentState);
end;

{ TSXSkinCustomControl }

procedure TSXSkinCustomControl.CreateParams(var Params:TCreateParams);
begin
 inherited;
 with Params do
  begin
   if not (csDesigning in ComponentState) then
    begin
     Style:=Style and not WS_CLIPCHILDREN;
     Style:=Style and not WS_CLIPSIBLINGS;
    end;
  end;
end;

function TSXSkinCustomControl.IsTransparent(X,Y:Integer;Limit:Integer=10):Boolean;
begin
 Result:=False;
end;

function TSXSkinCustomControl.CanShowControl:Boolean;
begin
 Result:=Visible or (csDesigning in ComponentState);
end;

procedure TSXSkinCustomControl.SetSkinLibrary(Value:TSXSkinLibrary);
begin
 if FSkinLibrary<>Value then
  begin
   if FSkinLibrary<>nil then
    begin
     FSkinLibrary.RemoveFreeNotification(Self);
     FSkinLibrary.RemoveControl(Self);
    end;
   FSkinLibrary:=Value;
   if FSkinLibrary<>nil then
    begin
     FSkinLibrary.FreeNotification(Self);
     FSkinLibrary.AddControl(Self);
    end;
   if not (csDestroying in ComponentState) then
    SkinChanged;
  end;
end;

procedure TSXSkinCustomControl.SetSkinStyle(const Value:String);
begin
 if FSkinStyle<>Value then
  begin
   FSkinStyle:=Value;
   SkinChanged;
  end;
end;

procedure TSXSkinCustomControl.SetParent(AParent:TWinControl);
var A:Integer;
   PC:TControl;
begin
 inherited;
 if Parent=nil then exit;
 if not (csLoading in ComponentState) and (csDesigning in ComponentState) and
    (SkinLibrary=nil) then
  begin
   PC:=Parent;
   repeat
    if (PC is TSXSkinCustomControl) and (TSXSkinCustomControl(PC).SkinLibrary<>nil) then
     SkinLibrary:=TSXSkinCustomControl(PC).SkinLibrary else
    if PC is TWinControl then
     begin
      for A:=0 to TWinControl(PC).ControlCount-1 do
       if (TWinControl(PC).Controls[A] is TSXSkinCustomControl) and
          (TSXSkinCustomControl(TWinControl(PC).Controls[A]).SkinLibrary<>nil) then
        begin
         SkinLibrary:=TSXSkinCustomControl(TWinControl(PC).Controls[A]).SkinLibrary;
         break;
        end;
      for A:=0 to TWinControl(PC).ComponentCount-1 do
       if TWinControl(PC).Components[A] is TSXSkinLibrary then
        begin
         SkinLibrary:=TSXSkinLibrary(TWinControl(PC).Components[A]);
         break;
        end;
     end;
    PC:=PC.Parent; 
   until (PC=nil) or not (PC is TWinControl) or (SkinLibrary<>nil);
  end;
end;

procedure TSXSkinCustomControl.Notification(AComponent:TComponent;Operation:TOperation);
begin
 inherited Notification(AComponent,Operation);
 if Operation=opRemove then
  begin
   if AComponent=FSkinLibrary then
    FSkinLibrary:=nil;
  end;
end;

procedure TSXSkinCustomControl.CMMouseEnter(var Msg: TMessage);
begin
 inherited;
 if Assigned(FOnMouseEnter) then FOnMouseEnter(Self);
end;

procedure TSXSkinCustomControl.CMMouseLeave(var Msg: TMessage);
begin
 inherited;
 if Assigned(FOnMouseLeave) then FOnMouseLeave(Self);
end;

procedure TSXSkinCustomControl.MouseDown(Button:TMouseButton;Shift:TShiftState;X,Y:Integer);
begin
 if TabStop and CanFocus then SetFocus;
 if Assigned(FOnMouseDown) then FOnMouseDown(Self,Button,Shift,X,Y);
end;

procedure TSXSkinCustomControl.MouseMove(Shift:TShiftState;X,Y:Integer);
begin
 inherited;
 if Assigned(FOnMouseMove) then FOnMouseMove(Self,Shift,X,Y);
end;

procedure TSXSkinCustomControl.MouseUp(Button:TMouseButton;Shift:TShiftState;X,Y:Integer);
begin
 if Assigned(FOnMouseUp) then FOnMouseUp(Self,Button,Shift,X,Y);
end;

procedure TSXSkinCustomControl.PaintRectToBitmap(DestCanvasHandle:HDC;
           DestCanvasRect:TRect;Rect:TRect;Rgn:HRGN;Bitmap:TBitmap32;X,Y:Integer;
           WithSubItems:Boolean);
var  A:Integer;
  R,R2:TRect;
     P:TPoint;
 XX,YY:Integer;
begin
 if WithSubItems then
  for A:=0 to ControlCount-1 do
   begin
    R2:=Controls[A].BoundsRect;
    if DoShowControl(Controls[A]) and RectInRegion(Rgn,R2) then
     begin
      IntersectRect(R,Rect,R2);
      if (Controls[A] is TSXSkinCustomControl) and (ControlsNotToPaint.IndexOf(Controls[A])<0) then
       begin
        XX:=R.Left-Rect.Left+X;
        YY:=R.Top-Rect.Top+Y;
        OffsetRect(R,-Controls[A].Left,-Controls[A].Top);
        OffsetRgn(Rgn,-Controls[A].Left,-Controls[A].Top);
        TSXSkinCustomControl(Controls[A]).PaintRectToBitmap(DestCanvasHandle,
           DestCanvasRect,R,Rgn,Bitmap,XX,YY,True);
        OffsetRgn(Rgn,Controls[A].Left,Controls[A].Top);
        if TSXSkinCustomControl(Controls[A]).HandleAllocated and
           not (csDesigning in ComponentState) then
         ValidateRect(TSXSkinCustomControl(Controls[A]).Handle,@R);
       end else
      if ((Controls[A] is TGraphicControl) or ((Controls[A] is TWinControl) and not (csCustomPaint in Controls[A].ControlState))) and
         (ControlsNotToPaint.IndexOf(Controls[A])<0) then
       begin
        XX:=X-Rect.Left+Controls[A].Left;
        YY:=Y-Rect.Top+Controls[A].Top;
        GetWindowOrgEx(Bitmap.Handle,P);
        SetWindowOrgEx(Bitmap.Handle,P.X-XX,P.Y-YY,nil);
        ControlsNotToPaint.Add(Controls[A]);
        try
         Controls[A].Perform(WM_PRINT,Integer(Bitmap.Handle),PRF_CHECKVISIBLE or PRF_CHILDREN or PRF_CLIENT or PRF_NONCLIENT or PRF_OWNED);
        finally
         ControlsNotToPaint.Extract(Controls[A]);
        end;
        SetWindowOrgEx(Bitmap.Handle,P.X,P.Y,nil);
        if Controls[A] is TWinControl then
         begin
          R:=Controls[A].BoundsRect;
          OffsetRect(R,-Controls[A].Left,-Controls[A].Top);
          if not (csDesigning in ComponentState) then
           ValidateRect(TWinControl(Controls[A]).Handle,@R);
         end;
       end;
     end;
   end;
end;

function TSXSkinCustomControl.NeedToPaintBackground:Boolean;
begin
 Result:=True;
end;

procedure TSXSkinCustomControl.WMPaint(var Msg:TWMPaint);
var BB:Boolean;
begin
 if not ((csDesigning in ComponentState) and (csPaintCopy in ControlState)) then
  begin
   if GetUpdateRgn(Handle,DrawRgn,False)=ERROR then
    begin
     DeleteObject(DrawRgn);
     DrawRgn:=CreateRectRgn(0,0,Width,Height);
    end;
   BB:=GetUpdateRect(Handle,DrawCR,False);
  end else BB:=False;
 if not BB then
  DrawCR:=Rect(0,0,Width,Height);
 DrawBR:=DrawCR;
 OffsetRect(DrawBR,Left,Top);
 inherited;
end;

procedure TSXSkinCustomControl.Paint;
const XorColor=$00FFD8CE;
var    B:TBitmap32;
   A,X,Y:Integer;
       R:TRect;
  TDrawCR:TRect;
  TDrawBR:TRect;
 TDrawRgn:HRGN;
       PC:TControl;
    DX,DY:Integer;
       BR:TRect;
   DoSibl:Boolean;
begin
 TDrawCR:=DrawCR;
 TDrawBR:=DrawBR;
 TDrawRgn:=DrawRgn;
 DrawRgn:=CreateRectRgn(0,0,Width,Height);
 DrawCR:=Rect(0,0,Width,Height);
 DrawBR:=BoundsRect;
 if not CanShowControl or (csLoading in ComponentState) or (csDestroying in ComponentState) then exit;
 if (csDesigning in ComponentState) and not (Self is TSXSkinPanel) and ((FSkinLibrary=nil) or not FSkinLibrary.Active) then
  begin
   DeleteObject(TDrawRgn);
   with Canvas do
    begin
     Pen.Style:=psDot;
     Pen.Mode:=pmXor;
     Pen.Color:=XorColor;
     Brush.Style:=bsClear;
     Rectangle(0,0,ClientWidth,ClientHeight);
    end;
   exit;
  end;
 if CanvasNotToPaint.IndexOf(Pointer(Canvas.Handle))>=0 then exit;
 B:=TBitmap32.Create;
 try
  B.DrawMode:=dmBlend;
  B.CombineMode:=cmBlend;
  B.SetSize(TDrawCR.Right-TDrawCR.Left,TDrawCR.Bottom-TDrawCR.Top);
  DoSibl:=True;
  if NeedToPaintBackground then
   begin
    PC:=Parent;
    BR:=TDrawBR;
    DX:=0; DY:=0;
    while (PC<>nil) and (PC is TSXSkinCustomControl) and (PC.Parent<>nil) and
          (PC.Parent is TSXSkinCustomControl) do
     begin
      Inc(DX,PC.Left);
      Inc(DY,PC.Top);
      OffsetRect(BR,PC.Left,PC.Top);
      OffsetRgn(TDrawRgn,PC.Left,PC.Top);
      PC:=PC.Parent;
     end;
    if (PC<>nil) and (PC is TSXSkinCustomControl) then
     begin
      OffsetRgn(TDrawRgn,Left,Top);
      Inc(DX,Left); Inc(DY,Top);
      if PC=Parent then
       TSXSkinCustomControl(PC).PaintRectToBitmap(Canvas.Handle,TDrawCR,BR,TDrawRgn,B,0,0,False) else
        begin
         DoSibl:=False;
         TSXSkinCustomControl(PC).PaintRectToBitmap(Canvas.Handle,TDrawCR,BR,TDrawRgn,B,0,0,True);
        end;
     end else B.Clear(Color32(Color));
    OffsetRgn(TDrawRgn,-DX,-DY);
   end;
  if DoSibl then
   begin
    OffsetRgn(TDrawRgn,Left,Top);
    for A:=0 to Parent.ControlCount-1 do
     if DoShowControl(Parent.Controls[A]) and (Parent.Controls[A] is TSXSkinCustomControl) and
        (ControlsNotToPaint.IndexOf(Parent.Controls[A])<0) then
      begin
       if Parent.Controls[A]=Self then
        begin
         OffsetRgn(TDrawRgn,-Left,-Top);
         PaintRectToBitmap(Canvas.Handle,TDrawCR,TDrawCR,TDrawRgn,B,0,0,True);
         OffsetRgn(TDrawRgn,Left,Top);
        end else
       if RectInRegion(TDrawRgn,Parent.Controls[A].BoundsRect) then
        begin
         IntersectRect(R,TDrawBR,Parent.Controls[A].BoundsRect);
         X:=R.Left-TDrawBR.Left;
         Y:=R.Top-TDrawBR.Top;
         OffsetRect(R,-Parent.Controls[A].Left,-Parent.Controls[A].Top);
         OffsetRgn(TDrawRgn,-Parent.Controls[A].Left,-Parent.Controls[A].Top);
         TSXSkinCustomControl(Parent.Controls[A]).PaintRectToBitmap(Canvas.Handle,TDrawCR,R,TDrawRgn,B,X,Y,True);
         OffsetRgn(TDrawRgn,Parent.Controls[A].Left,Parent.Controls[A].Top);
         if not (csDesigning in ComponentState) then
          ValidateRect(TSXSkinCustomControl(Parent.Controls[A]).Handle,@R);
        end;
      end;
    OffsetRgn(TDrawRgn,-Left,-Top);
   end;
{  for A:=0 to ControlCount-1 do
   if not (Controls[A] is TSXSkinCustomControl) then
    ExcludeClipRect(Canvas.Handle,Controls[A].Left,Controls[A].Top,
       Controls[A].Left+Controls[A].Width,Controls[A].Top+Controls[A].Height); }
 // ExcludeClipRect(Canvas.Handle,10,10,Width-10,Height-10);
  BitBlt(Canvas.Handle,TDrawCR.Left,TDrawCR.Top,B.Width,B.Height,B.Handle,0,0,SRCCOPY);
  if TestingRegions then
   begin
    Canvas.Brush.Style:=bsSolid;
    Canvas.Brush.Color:=RGB(random(256),random(256),random(256));
    FillRgn(Canvas.Handle,DrawRgn,Canvas.Brush.Handle);
   end;
 finally
  B.Free;
 end;
 DeleteObject(TDrawRgn);
end;

{procedure TSXSkinCustomControl.CMHintShow(var Message:TCMHintShow);
begin
 inherited;
 Message.Result:=0;
 Message.HintInfo.CursorRect:=Rect(0,0,Width,Height);
 Message.HintInfo.HintStr:=' ';
 Message.HintInfo.HintData:=FHintData;
 SetHintTimeout(Message.HintInfo^);
end;}

procedure TSXSkinCustomControl.WMEraseBkgnd(var Msg:TWmEraseBkgnd);
begin
 if csDesigning in ComponentState then
  begin
   Msg.Result:=1;
   inherited;
  end else Msg.Result:=1;
end;

procedure TSXSkinCustomControl.SetBounds(ALeft,ATop,AWidth,AHeight:Integer);
//var R:TRect;
begin
 if Parent=nil then
  begin
   inherited;
   exit;
  end;
 if CanShowControl and NeedToPaintBackground and ((Left<>ALeft) or (Top<>ATop) or (Width<>AWidth) or (Height<>AHeight)) then
  begin
   if csDesigning in ComponentState then
    begin
     inherited;
     Invalidate;
    end else
   {if Self is TSXSkinPanel then
    begin
     inherited;
     R:=Rect(0,0,Width,Height);
     ValidateRect(Handle,@R);
    end else}
     begin
      Visible:=False;
      inherited;
      Visible:=True;
     end;
  end else inherited;
end;

procedure TSXSkinCustomControl.SkinChanged;
var R:TRect;
begin
 if not (csLoading in ComponentState) then
  begin
   Realign;
   if HandleAllocated then
    begin
     R:=Rect(0,0,Width,Height);
     InvalidateRect(Handle,@R,False);
    end;
  end; 
end;

constructor TSXSkinCustomControl.Create(AOwner:TComponent);
begin
 inherited Create(AOwner);
 //FHintData:=TSXHintData.Create;
 ControlStyle:=ControlStyle+[csAcceptsControls,csCaptureMouse,csOpaque];
 DoubleBuffered:=False;
 TabStop:=False;
 DrawRgn:=CreateRectRgn(0,0,0,0);
end;

destructor TSXSkinCustomControl.Destroy;
begin
 DeleteObject(DrawRgn);
 SkinLibrary:=nil;
 //FHintData.Free;
 inherited Destroy;
end; 

initialization

 ControlsNotToPaint:=TList.Create;
 CanvasNotToPaint:=TList.Create;
 InitGetWindowRgnBox;

finalization

 ControlsNotToPaint.Free;
 CanvasNotToPaint.Free;

end.
