unit SXSkinImage;

////////////////////////////////////////////////////////////////////////////////
// SXSkinComponents: Skinnable Visual Controls with Alpha Channel             //
//----------------------------------------------------------------------------//
// Version: 1.0.1                                                             //
// Author: Alexey Sadovnikov                                                  //
// Web Site: http://www.saarixx.info                                          //
// E-Mail: mail@saarixx.info                                                  //
//----------------------------------------------------------------------------//
// LICENSE:                                                                   //
// 1. You can freely distribute this file.                                    //
// 2. You can not make any changes to this file.                              //
// 3. The only person who can change this file is Alexey Sadovnikov.          //
// 4. You can use this file in your freeware projects.                        //
// 5. If you want to use this file in your shareware project, you should      //
//    purchase a project license or a personal license of SXSkinComponents.   //
// 6. You can freely use, distribute and modify skins for SXSkinComponents.   //
// 7. You can create skins for SXSkinComponents.                              //
//----------------------------------------------------------------------------//
// Copyright 2006 by Alexey Sadovnikov. All Rights Reserved.                  //
////////////////////////////////////////////////////////////////////////////////

interface

{$I Compilers.inc}

uses Windows, Classes, SXSkinControl, GR32, SysUtils, Types, Messages, Controls;

type

  TSXSkinCustomImage=class(TSXSkinCustomControl)
   private
    FOnMouseEnter:TNotifyEvent;
    FOnMouseLeave:TNotifyEvent;
   protected
    function NeedToPaintBackground:Boolean; override;
    procedure CMMouseEnter(var Msg:TMessage); message CM_MOUSEENTER;
    procedure CMMouseLeave(var Msg:TMessage); message CM_MOUSELEAVE;
   public
    procedure PaintRectToBitmap(DestCanvasHandle:HDC;DestCanvasRect:TRect;
               Rect:TRect;Rgn:HRGN;Bitmap:TBitmap32;X,Y:Integer;
               WithSubItems:Boolean); override;   
    function IsTransparent(X,Y:Integer;Limit:Integer=10):Boolean; override;
   published
    property OnMouseEnter:TNotifyEvent read FOnMouseEnter write FOnMouseEnter;
    property OnMouseLeave:TNotifyEvent read FOnMouseLeave write FOnMouseLeave;
  end;

  TSXSkinImage=class(TSXSkinCustomImage)
   published
    property Align;
    property Anchors;
    property AutoSize;
    property Color;
    property Constraints;
    property Cursor;
    property DragCursor;
    property Font;
    //property HintData;
    property ParentColor;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property SkinLibrary;    
    property SkinStyle;
    property TabOrder;
    property TabStop;
    property Visible;
    property OnCanResize;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseEnter;
    property OnMouseLeave;
    property OnMouseMove;
    property OnMouseUp;
    property OnMouseWheel;
    property OnMouseWheelDown;
    property OnMouseWheelUp;
    property OnResize;
    property OnStartDrag;
  end;

implementation

uses SXSkinLibrary, Math;

{ TSXSkinCustomImage }

procedure TSXSkinCustomImage.PaintRectToBitmap(DestCanvasHandle:HDC;
           DestCanvasRect:TRect;Rect:TRect;Rgn:HRGN;Bitmap:TBitmap32;X,Y:Integer;
           WithSubItems:Boolean);
var  A,B,C,D:Integer;
       Style:TSXSkinGeneralStyle;
       Image:TSXSkinStyleImageElement;                
      Figure:TSXSkinStyleFigureElement;
     BoxTile:TSXSkinStyleBoxTileElement;
 DRect,SRect:TRect;
       TX,TY:Integer;
     TX2,TY2:Integer;
begin
 if (SkinLibrary<>nil) and SkinLibrary.CanBeUsed then
  begin
   A:=SkinLibrary.Styles.GetIndexByName(SkinStyle);
   if (A>=0) and (SkinLibrary.Styles[A] is TSXSkinGeneralStyle) then
    begin
     Style:=TSXSkinGeneralStyle(SkinLibrary.Styles[A]);
     for B:=0 to Style.Elements.Count-1 do
      if Style.Elements[B] is TSXSkinStyleImageElement then
       begin
        Image:=TSXSkinStyleImageElement(Style.Elements[B]);
        Image.ValidateBitmap(SkinLibrary);
        if (Image.Bitmap<>nil) and (Image.Bitmap.Width>0) and (Image.Bitmap.Height>0) then
         begin
          if EqualRect(Rect,Image.Bitmap.BoundsRect) then
           begin
            Image.Bitmap.DrawTo(Bitmap,X,Y);
           end else
            begin
             TX:=Rect.Left div Image.Bitmap.Width;
             TY:=Rect.Top div Image.Bitmap.Height;
             TX2:=Rect.Right div Image.Bitmap.Width;
             if Rect.Right mod Image.Bitmap.Width<>0 then Inc(TX2);
             TY2:=Rect.Bottom div Image.Bitmap.Height;
             if Rect.Bottom mod Image.Bitmap.Height<>0 then Inc(TY2);
             for C:=TY to TY2 do
              for D:=TX to TX2 do
               begin
                SRect:=Image.Bitmap.BoundsRect;
                DRect:=SRect;
                OffsetRect(DRect,D*Image.Bitmap.Width,C*Image.Bitmap.Height);
                if Rect.Left>DRect.Left then
                 begin
                  Inc(SRect.Left,Rect.Left-DRect.Left);
                  DRect.Left:=Rect.Left;
                 end;
                if Rect.Top>DRect.Top then
                 begin
                  Inc(SRect.Top,Rect.Top-DRect.Top);
                  DRect.Top:=Rect.Top;
                 end;
                if Rect.Right<DRect.Right then
                 begin
                  Dec(SRect.Right,DRect.Right-Rect.Right);
                  DRect.Right:=Rect.Right;
                 end;
                if Rect.Bottom<DRect.Bottom then
                 begin
                  Dec(SRect.Bottom,DRect.Bottom-Rect.Bottom);
                  DRect.Bottom:=Rect.Bottom;
                 end;
                OffsetRect(DRect,X-Rect.Left,Y-Rect.Top);
                Image.Bitmap.DrawTo(Bitmap,DRect,SRect);
               end;
            end;
         end;
       end else
      if Style.Elements[B] is TSXSkinStyleFigureElement then
       begin
        Figure:=TSXSkinStyleFigureElement(Style.Elements[B]);
        if Figure.FigureType=ssftSolidFill then
         begin
          DRect:=Types.Rect(X,Y,X+Rect.Right-Rect.Left,Y+Rect.Bottom-Rect.Top);
          Bitmap.FillRectTS(DRect,Figure.FillColor);
         end else
          begin
           Figure.EvalPrePaintParams(Self);
           Figure.DrawToBitmap(Bitmap,X-Rect.Left,Y-Rect.Top);
          end; 
       end else
      if Style.Elements[B] is TSXSkinStyleBoxTileElement then
       begin
        BoxTile:=TSXSkinStyleBoxTileElement(Style.Elements[B]);
        BoxTile.ValidateBitmaps(SkinLibrary);
        BoxTile.DrawToBitmap(Bitmap,X-Rect.Left,Y-Rect.Top,Width,Height,Rgn);
       end;
    end;
  end;
 inherited;
end;

function TSXSkinCustomImage.NeedToPaintBackground:Boolean;
var  A,B:Integer;
   Style:TSXSkinGeneralStyle;
   Image:TSXSkinStyleImageElement;
  Figure:TSXSkinStyleFigureElement;
 BoxTile:TSXSkinStyleBoxTileElement;
begin
 Result:=True;
 if (SkinLibrary<>nil) and SkinLibrary.CanBeUsed then
  begin
   A:=SkinLibrary.Styles.GetIndexByName(SkinStyle);
   if (A>=0) and (SkinLibrary.Styles[A] is TSXSkinGeneralStyle) then
    begin
     Style:=TSXSkinGeneralStyle(SkinLibrary.Styles[A]);
     Result:=Style.Transparent;
     if not Result then exit;
     for B:=0 to Style.Elements.Count-1 do
      if Style.Elements[B] is TSXSkinStyleImageElement then
       begin
        Image:=TSXSkinStyleImageElement(Style.Elements[B]);
        Result:=Image.Transparent;
        if not Result then exit;
       end else
      if Style.Elements[B] is TSXSkinStyleBoxTileElement then
       begin
        BoxTile:=TSXSkinStyleBoxTileElement(Style.Elements[B]);
        Result:=BoxTile.Transparent;
        if not Result then exit;
       end else
      if Style.Elements[B] is TSXSkinStyleFigureElement then
       begin
        Figure:=TSXSkinStyleFigureElement(Style.Elements[B]);
        Result:=Figure.FigureType<>ssftSolidFill;
        if not Result then exit;
       end;
    end;
  end;
end;

function TSXSkinCustomImage.IsTransparent(X,Y:Integer;Limit:Integer=10):Boolean;
var A,B,C:Integer;
    Style:TSXSkinGeneralStyle;
    Image:TSXSkinStyleImageElement;
   Figure:TSXSkinStyleFigureElement;
begin
 Result:=True;
 if (SkinLibrary<>nil) and SkinLibrary.CanBeUsed then
  begin
   A:=SkinLibrary.Styles.GetIndexByName(SkinStyle);
   if (A>=0) and (SkinLibrary.Styles[A] is TSXSkinGeneralStyle) then
    begin
     Style:=TSXSkinGeneralStyle(SkinLibrary.Styles[A]);
     for B:=0 to Style.Elements.Count-1 do
      begin
       if Style.Elements[B] is TSXSkinStyleImageElement then
        begin
         Image:=TSXSkinStyleImageElement(Style.Elements[B]);
         if AlphaComponent(Image.Bitmap.PixelF[X,Y])>Limit then
          Result:=False;
        end else
       if Style.Elements[B] is TSXSkinStyleFigureElement then
        begin
         Figure:=TSXSkinStyleFigureElement(Style.Elements[B]);
         case Figure.FigureType of
          ssftCircle:         ;
          ssftRectangle,
          ssftRoundRectangle: begin
                               C:=Figure.ControlsData.GetIndexByControl(Self);
                               if C>=0 then
                                Result:=not PtInRect(Figure.ControlsData[C].Rect,Point(X,Y));
                              end;
         end;
        end;
       if not Result then exit;
      end;
    end;
  end;
end;

procedure TSXSkinCustomImage.CMMouseEnter(var Msg:TMessage);
begin
 if Assigned(FOnMouseEnter) then
  FOnMouseEnter(Self);
end;

procedure TSXSkinCustomImage.CMMouseLeave(var Msg:TMessage);
begin
 if Assigned(FOnMouseLeave) then
  FOnMouseLeave(Self);
end;

end.
